@@ -1,4 +1,4 @@
-#!/usr/bin/env python2
+#!/usr/bin/env python3
 # Copyright 2016 The Emscripten Authors.  All rights reserved.
 # Emscripten is available under two separate licenses, the MIT license and the
 # University of Illinois/NCSA Open Source License.  Both these licenses can be
@@ -15,42 +15,38 @@ You can also use this for cmake and other configure-like
 stages. What happens is that all compilations done during
 this command are to native code, not JS, so that configure
 tests will work properly.
-
-Relevant defines:
-
-  CONFIGURE_CC - see emcc
 '''
 
 from __future__ import print_function
-import os, sys
+import sys
 from tools import shared
 from subprocess import CalledProcessError
 
+
 #
 # Main run() function
 #
 def run():
-  if len(sys.argv) < 2 or ('configure' not in sys.argv[1] and 'cmake' not in sys.argv[1]):
-    print('''
-  emconfigure is a helper for configure, setting various environment
-  variables so that emcc etc. are used. Typical usage:
+  if len(sys.argv) < 2 or sys.argv[1] in ('--version', '--help'):
+    print('''\
+emconfigure is a helper for configure, setting various environment
+variables so that emcc etc. are used. Typical usage:
 
-    emconfigure ./configure [FLAGS]
+  emconfigure ./configure [FLAGS]
 
-  (but you can run any command instead of configure)
+(but you can run any command instead of configure)''', file=sys.stderr)
+    return 1
 
-  ''', file=sys.stderr)
-  elif 'cmake' in sys.argv[1]:
-    node_js = shared.NODE_JS
-    if type(node_js) is list: node_js = node_js[0]
-    node_js = shared.Building.which(node_js)
-    node_js = node_js.replace('"', '\"')
-    sys.argv = sys.argv[:2] + ['-DCMAKE_CROSSCOMPILING_EMULATOR="' + node_js +'"'] + sys.argv[2:]
+  if 'cmake' in sys.argv[1]:
+    print('error: use `emcmake` rather then `emconfigure` for cmake projects', file=sys.stderr)
+    return 1
 
   try:
     shared.Building.configure(sys.argv[1:])
+    return 0
   except CalledProcessError as e:
-    sys.exit(e.returncode)
+    return e.returncode
+
 
 if __name__ == '__main__':
-  run()
+  sys.exit(run())
