@@ -1,4 +1,4 @@
-#!/usr/bin/env python2
+#!/usr/bin/env python3
 # Copyright 2016 The Emscripten Authors.  All rights reserved.
 # Emscripten is available under two separate licenses, the MIT license and the
 # University of Illinois/NCSA Open Source License.  Both these licenses can be
@@ -31,22 +31,23 @@ from subprocess import CalledProcessError
 # Main run() function
 #
 def run():
-  if len(sys.argv) < 2 or sys.argv[1] != 'make':
-    print('''
-  emmake is a helper for make, setting various environment
-  variables so that emcc etc. are used. Typical usage:
+  if len(sys.argv) < 2 or sys.argv[1] in ('--version', '--help'):
+    print('''\
+emmake is a helper for make, setting various environment
+variables so that emcc etc. are used. Typical usage:
 
-    emmake make [FLAGS]
+  emmake make [FLAGS]
 
-  (but you can run any command instead of make)
-
-  ''', file=sys.stderr)
+(but you can run any command instead of make)''', file=sys.stderr)
+    return 1
 
   try:
     shared.Building.make(sys.argv[1:])
   except CalledProcessError as e:
-    sys.exit(e.returncode)
+    return e.returncode
+
+  return 0
 
 
 if __name__ == '__main__':
-  run()
+  sys.exit(run())
