@@ -1,11 +1,35 @@
+/**
+ * @license
+ * Copyright 2015 The Emscripten Authors
+ * SPDX-License-Identifier: MIT
+ */
+
+#if BOOTSTRAPPING_STRUCT_INFO
+// When bootstrapping struct info, this is the entire library.  It is literally
+// just enough to run the bootstrap program that prints out C constants for us,
+// we obviously need to run without any such constants ourselves...
+assert(!LibraryManager.library);
+LibraryManager.library = {};
+#endif
+
 mergeInto(LibraryManager.library, {
+  $reallyNegative: function(x) {
+    return x < 0 || (x === 0 && (1/x) === -Infinity);
+  },
+
   // Performs printf-style formatting.
   //   format: A pointer to the format string.
   //   varargs: A pointer to the start of the arguments list.
   // Returns the resulting string string as a character array.
-  _formatString__deps: ['strlen', '_reallyNegative'],
-  _formatString: function(format, varargs) {
+  $formatString__deps: ['$reallyNegative', '$convertI32PairToI53', '$convertU32PairToI53'
+#if MINIMAL_RUNTIME
+    , '$intArrayFromString'
+#endif
+  ],
+  $formatString: function(format, varargs) {
+#if ASSERTIONS
     assert((varargs & 3) === 0);
+#endif
     var textIndex = format;
     var argIndex = varargs;
     // This must be called before reading a double or i64 vararg. It will bump the pointer properly.
@@ -14,11 +38,15 @@ mergeInto(LibraryManager.library, {
       if (type === 'double' || type === 'i64') {
         // move so the load is aligned
         if (ptr & 7) {
+#if ASSERTIONS
           assert((ptr & 7) === 4);
+#endif
           ptr += 4;
         }
       } else {
+#if ASSERTIONS
         assert((ptr & 3) === 0);
+#endif
       }
       return ptr;
     }
@@ -35,7 +63,9 @@ mergeInto(LibraryManager.library, {
                {{{ makeGetValue('argIndex', 4, 'i32', undefined, undefined, true, 4) }}}];
         argIndex += 8;
       } else {
+#if ASSERTIONS
         assert((argIndex & 3) === 0);
+#endif
         type = 'i32'; // varargs are always i32, i64, or double
         ret = {{{ makeGetValue('argIndex', 0, 'i32', undefined, undefined, true) }}};
         argIndex += 4;
@@ -169,13 +199,10 @@ mergeInto(LibraryManager.library, {
             var signed = next == {{{ charCode('d') }}} || next == {{{ charCode('i') }}};
             argSize = argSize || 4;
             currArg = getNextArg('i' + (argSize * 8));
-#if PRECISE_I64_MATH
-            var origArg = currArg;
-#endif
             var argText;
             // Flatten i64-1 [low, high] into a (slightly rounded) double
             if (argSize == 8) {
-              currArg = makeBigInt(currArg[0], currArg[1], next == {{{ charCode('u') }}});
+              currArg = next == {{{ charCode('u') }}} ? convertU32PairToI53(currArg[0], currArg[1]) : convertI32PairToI53(currArg[0], currArg[1]);
             }
             // Truncate to requested size.
             if (argSize <= 4) {
@@ -186,32 +213,14 @@ mergeInto(LibraryManager.library, {
             var currAbsArg = Math.abs(currArg);
             var prefix = '';
             if (next == {{{ charCode('d') }}} || next == {{{ charCode('i') }}}) {
-#if PRECISE_I64_MATH
-              if (argSize == 8 && typeof i64Math === 'object') argText = i64Math.stringify(origArg[0], origArg[1], null); else
-#endif
               argText = reSign(currArg, 8 * argSize, 1).toString(10);
             } else if (next == {{{ charCode('u') }}}) {
-#if PRECISE_I64_MATH
-              if (argSize == 8 && typeof i64Math === 'object') argText = i64Math.stringify(origArg[0], origArg[1], true); else
-#endif
               argText = unSign(currArg, 8 * argSize, 1).toString(10);
               currArg = Math.abs(currArg);
             } else if (next == {{{ charCode('o') }}}) {
               argText = (flagAlternative ? '0' : '') + currAbsArg.toString(8);
             } else if (next == {{{ charCode('x') }}} || next == {{{ charCode('X') }}}) {
               prefix = (flagAlternative && currArg != 0) ? '0x' : '';
-#if PRECISE_I64_MATH
-              if (argSize == 8 && typeof i64Math === 'object') {
-                if (origArg[1]) {
-                  argText = (origArg[1]>>>0).toString(16);
-                  var lower = (origArg[0]>>>0).toString(16);
-                  while (lower.length < 8) lower = '0' + lower;
-                  argText += lower;
-                } else {
-                  argText = (origArg[0]>>>0).toString(16);
-                }
-              } else
-#endif
               if (currArg < 0) {
                 // Represent negative numbers in hex as 2's complement.
                 currArg = -currArg;
@@ -316,7 +325,7 @@ mergeInto(LibraryManager.library, {
                 }
               } else if (next == {{{ charCode('f') }}} || next == {{{ charCode('F') }}}) {
                 argText = currArg.toFixed(effectivePrecision);
-                if (currArg === 0 && __reallyNegative(currArg)) {
+                if (currArg === 0 && reallyNegative(currArg)) {
                   argText = '-' + argText;
                 }
               }
@@ -434,12 +443,16 @@ mergeInto(LibraryManager.library, {
   },
 
   // printf/puts implementations for when musl is not pulled in - very partial. useful for tests, and when bootstrapping structInfo
-  printf__deps: ['_formatString'],
+  printf__deps: ['$formatString'
+#if MINIMAL_RUNTIME
+    , '$intArrayToString'
+#endif
+    ],
   printf: function(format, varargs) {
     // int printf(const char *restrict format, ...);
     // http://pubs.opengroup.org/onlinepubs/000095399/functions/printf.html
     // extra effort to support printf, even without a filesystem. very partial, very hackish
-    var result = __formatString(format, varargs);
+    var result = formatString(format, varargs);
     var string = intArrayToString(result);
     if (string[string.length-1] === '\n') string = string.substr(0, string.length-1); // remove a final \n, as Module.print will do that
     out(string);
@@ -447,7 +460,7 @@ mergeInto(LibraryManager.library, {
   },
   puts: function(s) {
     // extra effort to support puts, even without a filesystem. very partial, very hackish
-    var result = Pointer_stringify(s);
+    var result = UTF8ToString(s);
     var string = result.substr(0);
     if (string[string.length-1] === '\n') string = string.substr(0, string.length-1); // remove a final \n, as Module.print will do that
     out(string);
