@@ -1,10 +1,11 @@
-// Copyright 2013 The Emscripten Authors.  All rights reserved.
-// Emscripten is available under two separate licenses, the MIT license and the
-// University of Illinois/NCSA Open Source License.  Both these licenses can be
-// found in the LICENSE file.
+/**
+ * @license
+ * Copyright 2013 The Emscripten Authors
+ * SPDX-License-Identifier: MIT
+ */
 
 mergeInto(LibraryManager.library, {
-  $FS__deps: ['__setErrNo', '$PATH', '$PATH_FS', '$TTY', '$MEMFS',
+  $FS__deps: ['$setErrNo', '$PATH', '$PATH_FS', '$TTY', '$MEMFS',
 #if LibraryManager.has('library_idbfs.js')
     '$IDBFS',
 #endif
@@ -26,7 +27,59 @@ mergeInto(LibraryManager.library, {
     addAtInit('if (!Module["noFSInit"] && !FS.init.initialized) FS.init();');
     addAtMain('FS.ignorePermissions = false;');
     addAtExit('FS.quit();');
-    return 'FS.staticInit();' +
+    // We must statically create FS.FSNode here so that it is created in a manner
+    // that is visible to Closure compiler. That lets us use type annotations for
+    // Closure to the "this" pointer in various node creation functions.
+    return `var FSNode = /** @constructor */ function(parent, name, mode, rdev) {
+  if (!parent) {
+    parent = this;  // root node sets parent to itself
+  }
+  this.parent = parent;
+  this.mount = parent.mount;
+  this.mounted = null;
+  this.id = FS.nextInode++;
+  this.name = name;
+  this.mode = mode;
+  this.node_ops = {};
+  this.stream_ops = {};
+  this.rdev = rdev;
+};
+var readMode = 292/*{{{ cDefine("S_IRUGO") }}}*/ | 73/*{{{ cDefine("S_IXUGO") }}}*/;
+var writeMode = 146/*{{{ cDefine("S_IWUGO") }}}*/;
+Object.defineProperties(FSNode.prototype, {
+ read: {
+  get: /** @this{FSNode} */function() {
+   return (this.mode & readMode) === readMode;
+  },
+  set: /** @this{FSNode} */function(val) {
+   val ? this.mode |= readMode : this.mode &= ~readMode;
+  }
+ },
+ write: {
+  get: /** @this{FSNode} */function() {
+   return (this.mode & writeMode) === writeMode;
+  },
+  set: /** @this{FSNode} */function(val) {
+   val ? this.mode |= writeMode : this.mode &= ~writeMode;
+  }
+ },
+ isFolder: {
+  get: /** @this{FSNode} */function() {
+   return FS.isDir(this.mode);
+  }
+ },
+ isDevice: {
+  get: /** @this{FSNode} */function() {
+   return FS.isChrdev(this.mode);
+  }
+ }
+});
+FS.FSNode = FSNode;
+FS.staticInit();` +
+#if USE_CLOSURE_COMPILER
+           // Declare variable for Closure, FS.createPreloadedFile() below calls Browser.init()
+           '/**@suppress {duplicate, undefinedVars}*/var Browser;' +
+#endif
            // Get module methods from settings
            '{{{ EXPORTED_RUNTIME_METHODS.filter(function(func) { return func.substr(0, 3) === 'FS_' }).map(function(func){return 'Module["' + func + '"] = FS.' + func.substr(3) + ";"}).reduce(function(str, func){return str + func;}, '') }}}';
   },
@@ -58,7 +111,7 @@ mergeInto(LibraryManager.library, {
 
     handleFSError: function(e) {
       if (!(e instanceof FS.ErrnoError)) throw e + ' : ' + stackTrace();
-      return ___setErrNo(e.errno);
+      return setErrNo(e.errno);
     },
 
     //
@@ -179,9 +232,9 @@ mergeInto(LibraryManager.library, {
       }
     },
     lookupNode: function(parent, name) {
-      var err = FS.mayLookup(parent);
-      if (err) {
-        throw new FS.ErrnoError(err, parent);
+      var errCode = FS.mayLookup(parent);
+      if (errCode) {
+        throw new FS.ErrnoError(errCode, parent);
       }
       var hash = FS.hashName(parent.id, name);
 #if CASE_INSENSITIVE_FS
@@ -200,48 +253,6 @@ mergeInto(LibraryManager.library, {
       return FS.lookup(parent, name);
     },
     createNode: function(parent, name, mode, rdev) {
-      if (!FS.FSNode) {
-        FS.FSNode = function(parent, name, mode, rdev) {
-          if (!parent) {
-            parent = this;  // root node sets parent to itself
-          }
-          this.parent = parent;
-          this.mount = parent.mount;
-          this.mounted = null;
-          this.id = FS.nextInode++;
-          this.name = name;
-          this.mode = mode;
-          this.node_ops = {};
-          this.stream_ops = {};
-          this.rdev = rdev;
-        };
-
-        FS.FSNode.prototype = {};
-
-        // compatibility
-        var readMode = {{{ cDefine('S_IRUGO') }}} | {{{ cDefine('S_IXUGO') }}};
-        var writeMode = {{{ cDefine('S_IWUGO') }}};
-
-        // NOTE we must use Object.defineProperties instead of individual calls to
-        // Object.defineProperty in order to make closure compiler happy
-        Object.defineProperties(FS.FSNode.prototype, {
-          read: {
-            get: function() { return (this.mode & readMode) === readMode; },
-            set: function(val) { val ? this.mode |= readMode : this.mode &= ~readMode; }
-          },
-          write: {
-            get: function() { return (this.mode & writeMode) === writeMode; },
-            set: function(val) { val ? this.mode |= writeMode : this.mode &= ~writeMode; }
-          },
-          isFolder: {
-            get: function() { return FS.isDir(this.mode); }
-          },
-          isDevice: {
-            get: function() { return FS.isChrdev(this.mode); }
-          }
-        });
-      }
-
       var node = new FS.FSNode(parent, name, mode, rdev);
 
       FS.hashAddNode(node);
@@ -330,8 +341,8 @@ mergeInto(LibraryManager.library, {
       return 0;
     },
     mayLookup: function(dir) {
-      var err = FS.nodePermissions(dir, 'x');
-      if (err) return err;
+      var errCode = FS.nodePermissions(dir, 'x');
+      if (errCode) return errCode;
       if (!dir.node_ops.lookup) return {{{ cDefine('EACCES') }}};
       return 0;
     },
@@ -350,9 +361,9 @@ mergeInto(LibraryManager.library, {
       } catch (e) {
         return e.errno;
       }
-      var err = FS.nodePermissions(dir, 'wx');
-      if (err) {
-        return err;
+      var errCode = FS.nodePermissions(dir, 'wx');
+      if (errCode) {
+        return errCode;
       }
       if (isdir) {
         if (!FS.isDir(node.mode)) {
@@ -405,10 +416,8 @@ mergeInto(LibraryManager.library, {
     // SOCKFS is completed.
     createStream: function(stream, fd_start, fd_end) {
       if (!FS.FSStream) {
-        FS.FSStream = function(){};
-        FS.FSStream.prototype = {};
-        // compatibility
-        Object.defineProperties(FS.FSStream.prototype, {
+        FS.FSStream = /** @constructor */ function(){};
+        FS.FSStream.prototype = {
           object: {
             get: function() { return this.node; },
             set: function(val) { this.node = val; }
@@ -422,7 +431,7 @@ mergeInto(LibraryManager.library, {
           isAppend: {
             get: function() { return (this.flags & {{{ cDefine('O_APPEND') }}}); }
           }
-        });
+        };
       }
       // clone it, so we can return an instance of FSStream
       var newStream = new FS.FSStream();
@@ -504,25 +513,25 @@ mergeInto(LibraryManager.library, {
       FS.syncFSRequests++;
 
       if (FS.syncFSRequests > 1) {
-        console.log('warning: ' + FS.syncFSRequests + ' FS.syncfs operations in flight at once, probably just doing extra work');
+        err('warning: ' + FS.syncFSRequests + ' FS.syncfs operations in flight at once, probably just doing extra work');
       }
 
       var mounts = FS.getMounts(FS.root.mount);
       var completed = 0;
 
-      function doCallback(err) {
+      function doCallback(errCode) {
 #if ASSERTIONS
         assert(FS.syncFSRequests > 0);
 #endif
         FS.syncFSRequests--;
-        return callback(err);
+        return callback(errCode);
       }
 
-      function done(err) {
-        if (err) {
+      function done(errCode) {
+        if (errCode) {
           if (!done.errored) {
             done.errored = true;
-            return doCallback(err);
+            return doCallback(errCode);
           }
           return;
         }
@@ -641,9 +650,9 @@ mergeInto(LibraryManager.library, {
       if (!name || name === '.' || name === '..') {
         throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
       }
-      var err = FS.mayCreate(parent, name);
-      if (err) {
-        throw new FS.ErrnoError(err);
+      var errCode = FS.mayCreate(parent, name);
+      if (errCode) {
+        throw new FS.ErrnoError(errCode);
       }
       if (!parent.node_ops.mknod) {
         throw new FS.ErrnoError({{{ cDefine('EPERM') }}});
@@ -695,9 +704,9 @@ mergeInto(LibraryManager.library, {
         throw new FS.ErrnoError({{{ cDefine('ENOENT') }}});
       }
       var newname = PATH.basename(newpath);
-      var err = FS.mayCreate(parent, newname);
-      if (err) {
-        throw new FS.ErrnoError(err);
+      var errCode = FS.mayCreate(parent, newname);
+      if (errCode) {
+        throw new FS.ErrnoError(errCode);
       }
       if (!parent.node_ops.symlink) {
         throw new FS.ErrnoError({{{ cDefine('EPERM') }}});
@@ -749,17 +758,17 @@ mergeInto(LibraryManager.library, {
       }
       // we'll need to delete the old entry
       var isdir = FS.isDir(old_node.mode);
-      var err = FS.mayDelete(old_dir, old_name, isdir);
-      if (err) {
-        throw new FS.ErrnoError(err);
+      var errCode = FS.mayDelete(old_dir, old_name, isdir);
+      if (errCode) {
+        throw new FS.ErrnoError(errCode);
       }
       // need delete permissions if we'll be overwriting.
       // need create permissions if new doesn't already exist.
-      err = new_node ?
+      errCode = new_node ?
         FS.mayDelete(new_dir, new_name, isdir) :
         FS.mayCreate(new_dir, new_name);
-      if (err) {
-        throw new FS.ErrnoError(err);
+      if (errCode) {
+        throw new FS.ErrnoError(errCode);
       }
       if (!old_dir.node_ops.rename) {
         throw new FS.ErrnoError({{{ cDefine('EPERM') }}});
@@ -769,9 +778,9 @@ mergeInto(LibraryManager.library, {
       }
       // if we are going to change the parent, check write permissions
       if (new_dir !== old_dir) {
-        err = FS.nodePermissions(old_dir, 'w');
-        if (err) {
-          throw new FS.ErrnoError(err);
+        errCode = FS.nodePermissions(old_dir, 'w');
+        if (errCode) {
+          throw new FS.ErrnoError(errCode);
         }
       }
       try {
@@ -779,7 +788,7 @@ mergeInto(LibraryManager.library, {
           FS.trackingDelegate['willMovePath'](old_path, new_path);
         }
       } catch(e) {
-        console.log("FS.trackingDelegate['willMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
+        err("FS.trackingDelegate['willMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
       }
       // remove the node from the lookup hash
       FS.hashRemoveNode(old_node);
@@ -796,7 +805,7 @@ mergeInto(LibraryManager.library, {
       try {
         if (FS.trackingDelegate['onMovePath']) FS.trackingDelegate['onMovePath'](old_path, new_path);
       } catch(e) {
-        console.log("FS.trackingDelegate['onMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
+        err("FS.trackingDelegate['onMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
       }
     },
     rmdir: function(path) {
@@ -804,9 +813,9 @@ mergeInto(LibraryManager.library, {
       var parent = lookup.node;
       var name = PATH.basename(path);
       var node = FS.lookupNode(parent, name);
-      var err = FS.mayDelete(parent, name, true);
-      if (err) {
-        throw new FS.ErrnoError(err);
+      var errCode = FS.mayDelete(parent, name, true);
+      if (errCode) {
+        throw new FS.ErrnoError(errCode);
       }
       if (!parent.node_ops.rmdir) {
         throw new FS.ErrnoError({{{ cDefine('EPERM') }}});
@@ -819,14 +828,14 @@ mergeInto(LibraryManager.library, {
           FS.trackingDelegate['willDeletePath'](path);
         }
       } catch(e) {
-        console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
+        err("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
       }
       parent.node_ops.rmdir(parent, name);
       FS.destroyNode(node);
       try {
         if (FS.trackingDelegate['onDeletePath']) FS.trackingDelegate['onDeletePath'](path);
       } catch(e) {
-        console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
+        err("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
       }
     },
     readdir: function(path) {
@@ -842,12 +851,12 @@ mergeInto(LibraryManager.library, {
       var parent = lookup.node;
       var name = PATH.basename(path);
       var node = FS.lookupNode(parent, name);
-      var err = FS.mayDelete(parent, name, false);
-      if (err) {
+      var errCode = FS.mayDelete(parent, name, false);
+      if (errCode) {
         // According to POSIX, we should map EISDIR to EPERM, but
         // we instead do what Linux does (and we must, as we use
         // the musl linux libc).
-        throw new FS.ErrnoError(err);
+        throw new FS.ErrnoError(errCode);
       }
       if (!parent.node_ops.unlink) {
         throw new FS.ErrnoError({{{ cDefine('EPERM') }}});
@@ -860,14 +869,14 @@ mergeInto(LibraryManager.library, {
           FS.trackingDelegate['willDeletePath'](path);
         }
       } catch(e) {
-        console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
+        err("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
       }
       parent.node_ops.unlink(parent, name);
       FS.destroyNode(node);
       try {
         if (FS.trackingDelegate['onDeletePath']) FS.trackingDelegate['onDeletePath'](path);
       } catch(e) {
-        console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
+        err("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
       }
     },
     readlink: function(path) {
@@ -967,9 +976,9 @@ mergeInto(LibraryManager.library, {
       if (!FS.isFile(node.mode)) {
         throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
       }
-      var err = FS.nodePermissions(node, 'w');
-      if (err) {
-        throw new FS.ErrnoError(err);
+      var errCode = FS.nodePermissions(node, 'w');
+      if (errCode) {
+        throw new FS.ErrnoError(errCode);
       }
       node.node_ops.setattr(node, {
         size: len,
@@ -1047,9 +1056,9 @@ mergeInto(LibraryManager.library, {
       // create and write to a file with read-only permissions; it is read-only
       // for later use)
       if (!created) {
-        var err = FS.mayOpen(node, flags);
-        if (err) {
-          throw new FS.ErrnoError(err);
+        var errCode = FS.mayOpen(node, flags);
+        if (errCode) {
+          throw new FS.ErrnoError(errCode);
         }
       }
       // do truncation if necessary
@@ -1057,7 +1066,7 @@ mergeInto(LibraryManager.library, {
         FS.truncate(node, 0);
       }
       // we've already handled these, don't pass down to the underlying vfs
-      flags &= ~({{{ cDefine('O_EXCL') }}} | {{{ cDefine('O_TRUNC') }}});
+      flags &= ~({{{ cDefine('O_EXCL') }}} | {{{ cDefine('O_TRUNC') }}} | {{{ cDefine('O_NOFOLLOW') }}});
 
       // register the stream with the filesystem
       var stream = FS.createStream({
@@ -1079,7 +1088,7 @@ mergeInto(LibraryManager.library, {
         if (!FS.readFiles) FS.readFiles = {};
         if (!(path in FS.readFiles)) {
           FS.readFiles[path] = 1;
-          console.log("FS.trackingDelegate error on read file: " + path);
+          err("FS.trackingDelegate error on read file: " + path);
         }
       }
       try {
@@ -1094,7 +1103,7 @@ mergeInto(LibraryManager.library, {
           FS.trackingDelegate['onOpenFile'](path, trackingFlags);
         }
       } catch(e) {
-        console.log("FS.trackingDelegate['onOpenFile']('"+path+"', flags) threw an exception: " + e.message);
+        err("FS.trackingDelegate['onOpenFile']('"+path+"', flags) threw an exception: " + e.message);
       }
       return stream;
     },
@@ -1132,6 +1141,9 @@ mergeInto(LibraryManager.library, {
       return stream.position;
     },
     read: function(stream, buffer, offset, length, position) {
+#if CAN_ADDRESS_2GB
+      offset >>>= 0;
+#endif
       if (length < 0 || position < 0) {
         throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
       }
@@ -1158,6 +1170,9 @@ mergeInto(LibraryManager.library, {
       return bytesRead;
     },
     write: function(stream, buffer, offset, length, position, canOwn) {
+#if CAN_ADDRESS_2GB
+      offset >>>= 0;
+#endif
       if (length < 0 || position < 0) {
         throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
       }
@@ -1173,7 +1188,7 @@ mergeInto(LibraryManager.library, {
       if (!stream.stream_ops.write) {
         throw new FS.ErrnoError({{{ cDefine('EINVAL') }}});
       }
-      if (stream.flags & {{{ cDefine('O_APPEND') }}}) {
+      if (stream.seekable && stream.flags & {{{ cDefine('O_APPEND') }}}) {
         // seek to the end before writing in append mode
         FS.llseek(stream, 0, {{{ cDefine('SEEK_END') }}});
       }
@@ -1188,7 +1203,7 @@ mergeInto(LibraryManager.library, {
       try {
         if (stream.path && FS.trackingDelegate['onWriteToFile']) FS.trackingDelegate['onWriteToFile'](stream.path);
       } catch(e) {
-        console.log("FS.trackingDelegate['onWriteToFile']('"+stream.path+"') threw an exception: " + e.message);
+        err("FS.trackingDelegate['onWriteToFile']('"+stream.path+"') threw an exception: " + e.message);
       }
       return bytesWritten;
     },
@@ -1211,6 +1226,9 @@ mergeInto(LibraryManager.library, {
       stream.stream_ops.allocate(stream, offset, length);
     },
     mmap: function(stream, buffer, offset, length, position, prot, flags) {
+#if CAN_ADDRESS_2GB
+      offset >>>= 0;
+#endif
       // User requests writing to file (prot & PROT_WRITE != 0).
       // Checking if we have permissions to write to the file unless
       // MAP_PRIVATE flag is set. According to POSIX spec it is possible
@@ -1231,6 +1249,9 @@ mergeInto(LibraryManager.library, {
       return stream.stream_ops.mmap(stream, buffer, offset, length, position, prot, flags);
     },
     msync: function(stream, buffer, offset, length, mmapFlags) {
+#if CAN_ADDRESS_2GB
+      offset >>>= 0;
+#endif
       if (!stream || !stream.stream_ops.msync) {
         return 0;
       }
@@ -1296,9 +1317,9 @@ mergeInto(LibraryManager.library, {
       if (!FS.isDir(lookup.node.mode)) {
         throw new FS.ErrnoError({{{ cDefine('ENOTDIR') }}});
       }
-      var err = FS.nodePermissions(lookup.node, 'x');
-      if (err) {
-        throw new FS.ErrnoError(err);
+      var errCode = FS.nodePermissions(lookup.node, 'x');
+      if (errCode) {
+        throw new FS.ErrnoError(errCode);
       }
       FS.currentPath = lookup.path;
     },
@@ -1421,9 +1442,9 @@ mergeInto(LibraryManager.library, {
     },
     ensureErrnoError: function() {
       if (FS.ErrnoError) return;
-      FS.ErrnoError = function ErrnoError(errno, node) {
+      FS.ErrnoError = /** @this{Object} */ function ErrnoError(errno, node) {
         this.node = node;
-        this.setErrno = function(errno) {
+        this.setErrno = /** @this{Object} */ function(errno) {
           this.errno = errno;
 #if ASSERTIONS
           for (var key in ERRNO_CODES) {
@@ -1538,7 +1559,7 @@ mergeInto(LibraryManager.library, {
       if (ret.exists) {
         return ret.object;
       } else {
-        ___setErrNo(ret.error);
+        setErrNo(ret.error);
         return null;
       }
     },
@@ -1693,7 +1714,7 @@ mergeInto(LibraryManager.library, {
       } else {
         throw new Error('Cannot load without read() or XMLHttpRequest.');
       }
-      if (!success) ___setErrNo({{{ cDefine('EIO') }}});
+      if (!success) setErrNo({{{ cDefine('EIO') }}});
       return success;
     },
     // Creates a file record for lazy-loading from a URL. XXX This requires a synchronous
@@ -1701,11 +1722,12 @@ mergeInto(LibraryManager.library, {
     // either --preload-file in emcc or FS.createPreloadedFile
     createLazyFile: function(parent, name, url, canRead, canWrite) {
       // Lazy chunked Uint8Array (implements get and length from Uint8Array). Actual getting is abstracted away for eventual reuse.
+      /** @constructor */
       function LazyUint8Array() {
         this.lengthKnown = false;
         this.chunks = []; // Loaded chunks. Index is the chunk number
       }
-      LazyUint8Array.prototype.get = function LazyUint8Array_get(idx) {
+      LazyUint8Array.prototype.get = /** @this{Object} */ function LazyUint8Array_get(idx) {
         if (idx > this.length-1 || idx < 0) {
           return undefined;
         }
@@ -1754,7 +1776,7 @@ mergeInto(LibraryManager.library, {
           xhr.send(null);
           if (!(xhr.status >= 200 && xhr.status < 300 || xhr.status === 304)) throw new Error("Couldn't load " + url + ". Status: " + xhr.status);
           if (xhr.response !== undefined) {
-            return new Uint8Array(xhr.response || []);
+            return new Uint8Array(/** @type{Array<number>} */(xhr.response || []));
           } else {
             return intArrayFromString(xhr.responseText || '', true);
           }
@@ -1776,7 +1798,7 @@ mergeInto(LibraryManager.library, {
           chunkSize = datalength = 1; // this will force getter(0)/doXHR do download the whole file
           datalength = this.getter(0).length;
           chunkSize = datalength;
-          console.log("LazyFiles on gzip forces download of the whole file when length is accessed");
+          out("LazyFiles on gzip forces download of the whole file when length is accessed");
         }
 
         this._length = datalength;
@@ -1788,7 +1810,7 @@ mergeInto(LibraryManager.library, {
         var lazyArray = new LazyUint8Array();
         Object.defineProperties(lazyArray, {
           length: {
-            get: function() {
+            get: /** @this{Object} */ function() {
               if(!this.lengthKnown) {
                 this.cacheLength();
               }
@@ -1796,7 +1818,7 @@ mergeInto(LibraryManager.library, {
             }
           },
           chunkSize: {
-            get: function() {
+            get: /** @this{Object} */ function() {
               if(!this.lengthKnown) {
                 this.cacheLength();
               }
@@ -1823,7 +1845,7 @@ mergeInto(LibraryManager.library, {
       // Add a function that defers querying the file size until it is asked the first time.
       Object.defineProperties(node, {
         usedBytes: {
-          get: function() { return this.contents.length; }
+          get: /** @this {FSNode} */ function() { return this.contents.length; }
         }
       });
       // override each stream op with one that tries to force load the lazy file first
@@ -1938,7 +1960,7 @@ mergeInto(LibraryManager.library, {
         return onerror(e);
       }
       openRequest.onupgradeneeded = function openRequest_onupgradeneeded() {
-        console.log('creating db');
+        out('creating db');
         var db = openRequest.result;
         db.createObjectStore(FS.DB_STORE_NAME);
       };
