@@ -1,67 +1,36 @@
-// Copyright 2015 The Emscripten Authors.  All rights reserved.
-// Emscripten is available under two separate licenses, the MIT license and the
-// University of Illinois/NCSA Open Source License.  Both these licenses can be
-// found in the LICENSE file.
+/**
+ * @license
+ * Copyright 2015 The Emscripten Authors
+ * SPDX-License-Identifier: MIT
+ */
 
 var LibraryPThreadStub = {
   // ===================================================================================
   // Stub implementation for pthread.h when not compiling with pthreads support enabled.
   // ===================================================================================
 
-  emscripten_has_threading_support: function() {
-    return 0;
-  },
-
-  emscripten_num_logical_cores: function() {
-    return 1;
-  },
-
-  emscripten_force_num_logical_cores: function(cores) {
-    // Ignored, no threading available.
-  },
-
-  emscripten_is_main_runtime_thread: function() {
-    return 1;
-  },
-
+#if ENVIRONMENT_MAY_BE_WORKER || WASM_BACKEND
   emscripten_is_main_browser_thread: function() {
+#if MINIMAL_RUNTIME
+    return typeof importScripts === 'undefined';
+#else
     return !ENVIRONMENT_IS_WORKER;
+#endif
   },
-
-  emscripten_main_thread_process_queued_calls: function() {
-    // We will never have any queued calls to process, so no-op.
+#else
+  emscripten_is_main_browser_thread__asm: true,
+  emscripten_is_main_browser_thread__sig: 'i',
+  emscripten_is_main_browser_thread: function() {
+    return 1;
   },
+#endif
 
-  pthread_barrier_init: function() {},
-  pthread_barrier_wait: function() {},
-  pthread_barrier_destroy: function() {},
-  pthread_mutex_init: function() {},
-  pthread_mutex_destroy: function() {},
   pthread_mutexattr_init: function() {},
   pthread_mutexattr_setschedparam: function() {},
   pthread_mutexattr_setprotocol: function() {},
   pthread_mutexattr_settype: function() {},
   pthread_mutexattr_destroy: function() {},
 
-  pthread_mutex_lock__asm: true,
-  pthread_mutex_lock__sig: 'ii',
-  pthread_mutex_lock: function(x) {
-    x = x | 0;
-    return 0;
-  },
-  pthread_mutex_unlock__asm: true,
-  pthread_mutex_unlock__sig: 'ii',
-  pthread_mutex_unlock: function(x) {
-    x = x | 0;
-    return 0;
-  },
-  pthread_mutex_trylock__asm: true,
-  pthread_mutex_trylock__sig: 'ii',
-  pthread_mutex_trylock: function(x) {
-    x = x | 0;
-    return 0;
-  },
-
   pthread_mutexattr_setpshared: function(attr, pshared) {
     // XXX implement if/when getpshared is required
     return 0;
@@ -69,9 +38,7 @@ var LibraryPThreadStub = {
 
   pthread_cond_init: function() { return 0; },
   pthread_cond_destroy: function() { return 0; },
-  pthread_cond_wait: function() { return 0; },
   pthread_cond_timedwait: function() { return 0; },
-  pthread_cond_signal: function() { return 0; },
 
   pthread_condattr_init: function() { return 0; },
   pthread_condattr_destroy: function() { return 0; },
@@ -80,13 +47,6 @@ var LibraryPThreadStub = {
   pthread_condattr_getclock: function() { return 0; },
   pthread_condattr_getpshared: function() { return 0; },
 
-  pthread_cond_broadcast__asm: true,
-  pthread_cond_broadcast__sig: 'ii',
-  pthread_cond_broadcast: function(x) {
-    x = x | 0;
-    return 0;
-  },
-
   pthread_attr_init: function(attr) {
     /* int pthread_attr_init(pthread_attr_t *attr); */
     //FIXME: should allocate a pthread_attr_t
@@ -111,55 +71,16 @@ var LibraryPThreadStub = {
     {{{ makeSetValue('stacksize', '0', 'TOTAL_STACK', 'i32') }}};
     return 0;
   },
+  pthread_attr_getdetachstate: function(attr, detachstate) {
+    /* int pthread_attr_getdetachstate(const pthread_attr_t *attr, int *detachstate); */
+    return 0;
+  },
 
   pthread_setcancelstate: function() { return 0; },
-
-  pthread_once: function(ptr, func) {
-    if (!_pthread_once.seen) _pthread_once.seen = {};
-    if (ptr in _pthread_once.seen) return;
-    Module['dynCall_v'](func);
-    _pthread_once.seen[ptr] = 1;
-  },
-
-  $PTHREAD_SPECIFIC: {},
-  $PTHREAD_SPECIFIC_NEXT_KEY: 1,
-  pthread_key_create__deps: ['$PTHREAD_SPECIFIC', '$PTHREAD_SPECIFIC_NEXT_KEY', '$ERRNO_CODES'],
-  pthread_key_create: function(key, destructor) {
-    if (key == 0) {
-      return ERRNO_CODES.EINVAL;
-    }
-    {{{ makeSetValue('key', '0', 'PTHREAD_SPECIFIC_NEXT_KEY', 'i32*') }}};
-    // values start at 0
-    PTHREAD_SPECIFIC[PTHREAD_SPECIFIC_NEXT_KEY] = 0;
-    PTHREAD_SPECIFIC_NEXT_KEY++;
-    return 0;
-  },
-
-  pthread_getspecific__deps: ['$PTHREAD_SPECIFIC'],
-  pthread_getspecific: function(key) {
-    return PTHREAD_SPECIFIC[key] || 0;
-  },
-
-  pthread_setspecific__deps: ['$PTHREAD_SPECIFIC', '$ERRNO_CODES'],
-  pthread_setspecific: function(key, value) {
-    if (!(key in PTHREAD_SPECIFIC)) {
-      return ERRNO_CODES.EINVAL;
-    }
-    PTHREAD_SPECIFIC[key] = value;
-    return 0;
-  },
-
-  pthread_key_delete__deps: ['$PTHREAD_SPECIFIC', '$ERRNO_CODES'],
-  pthread_key_delete: function(key) {
-    if (key in PTHREAD_SPECIFIC) {
-      delete PTHREAD_SPECIFIC[key];
-      return 0;
-    }
-    return ERRNO_CODES.EINVAL;
-  },
+  pthread_setcanceltype: function() { return 0; },
 
   pthread_cleanup_push: function(routine, arg) {
-    __ATEXIT__.push(function() { Module['dynCall_vi'](routine, arg) })
+    __ATEXIT__.push(function() { {{{ makeDynCall('vi') }}}(routine, arg) })
     _pthread_cleanup_push.level = __ATEXIT__.length;
   },
 
@@ -169,10 +90,17 @@ var LibraryPThreadStub = {
     _pthread_cleanup_push.level = __ATEXIT__.length;
   },
 
+  _pthread_cleanup_push__sig: 'vii',
   _pthread_cleanup_push: 'pthread_cleanup_push',
+
+  _pthread_cleanup_pop__sig: 'v',
   _pthread_cleanup_pop: 'pthread_cleanup_pop',
 
-  pthread_sigmask: function() { return 0; },
+  // pthread_sigmask - examine and change mask of blocked signals
+  pthread_sigmask: function(how, set, oldset) {
+    err('pthread_sigmask() is not supported: this is a no-op.');
+    return 0;
+  },
 
   pthread_rwlock_init: function() { return 0; },
   pthread_rwlock_destroy: function() { return 0; },
@@ -199,7 +127,7 @@ var LibraryPThreadStub = {
   pthread_attr_setschedparam: function() {},
   pthread_attr_setstacksize: function() {},
 
-  pthread_create: function() {
+  {{{ USE_LSAN || USE_ASAN ? 'emscripten_builtin_' : '' }}}pthread_create: function() {
     return {{{ cDefine('EAGAIN') }}};
   },
   pthread_cancel: function() {},
@@ -209,7 +137,7 @@ var LibraryPThreadStub = {
   },
 
   pthread_equal: function(x, y) { return x == y },
-  pthread_join: function() {},
+  {{{ USE_LSAN ? 'emscripten_builtin_' : '' }}}pthread_join: function() {},
   pthread_detach: function() {},
 
   sem_init: function() {},
@@ -222,30 +150,30 @@ var LibraryPThreadStub = {
   emscripten_main_browser_thread_id: function() { return _pthread_self(); },
 
   // When pthreads is not enabled, we can't use the Atomics futex api to do proper sleeps, so simulate a busy spin wait loop instead.
+  usleep__deps: ['emscripten_get_now'],
   usleep: function(useconds) {
     // int usleep(useconds_t useconds);
     // http://pubs.opengroup.org/onlinepubs/000095399/functions/usleep.html
     // We're single-threaded, so use a busy loop. Super-ugly.
-    var msec = useconds / 1000;
-    if ((ENVIRONMENT_IS_WEB || ENVIRONMENT_IS_WORKER) && self['performance'] && self['performance']['now']) {
-      var start = self['performance']['now']();
-      while (self['performance']['now']() - start < msec) {
-        // Do nothing.
-      }
-    } else {
-      var start = Date.now();
-      while (Date.now() - start < msec) {
-        // Do nothing.
-      }
+    var start = _emscripten_get_now();
+    while (_emscripten_get_now() - start < useconds / 1000) {
+      // Do nothing.
     }
-    return 0;
   },
 
-  nanosleep__deps: ['usleep'],
+  nanosleep__deps: ['usleep', '$setErrNo'],
   nanosleep: function(rqtp, rmtp) {
     // int nanosleep(const struct timespec  *rqtp, struct timespec *rmtp);
+    if (rqtp === 0) {
+      setErrNo({{{ cDefine('EINVAL') }}});
+      return -1;
+    }
     var seconds = {{{ makeGetValue('rqtp', C_STRUCTS.timespec.tv_sec, 'i32') }}};
     var nanoseconds = {{{ makeGetValue('rqtp', C_STRUCTS.timespec.tv_nsec, 'i32') }}};
+    if (nanoseconds < 0 || nanoseconds > 999999999 || seconds < 0) {
+      setErrNo({{{ cDefine('EINVAL') }}});
+      return -1;
+    }
     if (rmtp !== 0) {
       {{{ makeSetValue('rmtp', C_STRUCTS.timespec.tv_sec, '0', 'i32') }}};
       {{{ makeSetValue('rmtp', C_STRUCTS.timespec.tv_nsec, '0', 'i32') }}};
@@ -343,15 +271,37 @@ var LibraryPThreadStub = {
     {{{ makeStructuralReturn(['l', 'h']) }}};
   },
 
+  emscripten_atomic_add_u32__sig: 'iii',
   emscripten_atomic_add_u32: 'llvm_atomic_load_add_i32_p0i32',
+
+  emscripten_atomic_load_u64__sig: 'iii',
   emscripten_atomic_load_u64: '__atomic_load_8',
+
+  emscripten_atomic_store_u64__sig: 'viiii',
   emscripten_atomic_store_u64: '__atomic_store_8',
+
+  emscripten_atomic_cas_u64__sig: 'iiiiiiii',
   emscripten_atomic_cas_u64: '__atomic_compare_exchange_8',
+
+  emscripten_atomic_exchange_u64__sig: 'iiiii',
   emscripten_atomic_exchange_u64: '__atomic_exchange_8',
+
+  _emscripten_atomic_fetch_and_add_u64__sig: 'iiiii',
   _emscripten_atomic_fetch_and_add_u64: '__atomic_fetch_add_8',
+
+  _emscripten_atomic_fetch_and_add_u64__sig: 'iiiii',
+  _emscripten_atomic_fetch_and_add_u64: '__atomic_fetch_add_8',
+
+  _emscripten_atomic_fetch_and_sub_u64__sig: 'iiiii',
   _emscripten_atomic_fetch_and_sub_u64: '__atomic_fetch_sub_8',
+
+  _emscripten_atomic_fetch_and_and_u64__sig: 'iiiii',
   _emscripten_atomic_fetch_and_and_u64: '__atomic_fetch_and_8',
+
+  _emscripten_atomic_fetch_and_or_u64__sig: 'iiiii',
   _emscripten_atomic_fetch_and_or_u64: '__atomic_fetch_or_8',
+
+  _emscripten_atomic_fetch_and_xor_u64__sig: 'iiiii',
   _emscripten_atomic_fetch_and_xor_u64: '__atomic_fetch_xor_8',
 
   __wait: function() {},
