@@ -1,3 +1,9 @@
+/**
+ * @license
+ * Copyright 2018 The Emscripten Authors
+ * SPDX-License-Identifier: MIT
+ */
+
 var LibraryWebSocket = {
   $WS: {
     sockets: [null],
@@ -109,6 +115,7 @@ var LibraryWebSocket = {
 #endif
       return {{{ cDefine('EMSCRIPTEN_RESULT_INVALID_TARGET') }}};
     }
+    HEAPU32[urlLength>>2] = lengthBytesUTF8(socket.url);
     return {{{ cDefine('EMSCRIPTEN_RESULT_SUCCESS') }}};
   },
 
