@@ -1,13 +1,12 @@
-// Copyright 2010 The Emscripten Authors.  All rights reserved.
-// Emscripten is available under two separate licenses, the MIT license and the
-// University of Illinois/NCSA Open Source License.  Both these licenses can be
-// found in the LICENSE file.
-//
+/**
+ * @license
+ * Copyright 2010 The Emscripten Authors
+ * SPDX-License-Identifier: MIT
+ */
+
 // Various tools for parsing LLVM. Utilities of various sorts, that are
 // specific to Emscripten (and hence not in utility.js).
 
-//"use strict";
-
 // Does simple 'macro' substitution, using Django-like syntax,
 // {{{ code }}} will be replaced with |eval(code)|.
 // NOTE: Be careful with that ret check. If ret is |0|, |ret ? ret.toString() : ''| would result in ''!
@@ -1320,11 +1319,13 @@ function makeStructuralReturn(values, inAsm) {
 }
 
 function makeThrow(what) {
-  if (ASSERTIONS) {
-    return 'throw ' + what + (DISABLE_EXCEPTION_CATCHING == 1 ? ' + " - Exception catching is disabled, this exception cannot be caught. Compile with -s DISABLE_EXCEPTION_CATCHING=0 or DISABLE_EXCEPTION_CATCHING=2 to catch."' : '') + ';';
-  } else {
-  return 'throw ' + what + ';';
+  if (ASSERTIONS && DISABLE_EXCEPTION_CATCHING == 1) {
+    what += ' + " - Exception catching is disabled, this exception cannot be caught. Compile with -s DISABLE_EXCEPTION_CATCHING=0 or DISABLE_EXCEPTION_CATCHING=2 to catch."';
+    if (MAIN_MODULE) {
+      what += ' + " (note: in dynamic linking, if a side module wants exceptions, the main module must be built with that support)"';
+    }
   }
+  return 'throw ' + what + ';';
 }
 
 function makeSignOp(value, type, op, force, ignore) {
@@ -1549,19 +1550,37 @@ function makeRetainedCompilerSettings() {
 var WASM_PAGE_SIZE = 65536;
 var ASMJS_PAGE_SIZE = 16777216;
 
-function getPageSize() {
+function getMemoryPageSize() {
   return WASM ? WASM_PAGE_SIZE : ASMJS_PAGE_SIZE;
 }
 
+// Page size reported by some POSIX calls, mostly filesystem. This does not
+// depend on the memory page size which differs between wasm and asm.js, and
+// makes us report a consistent value despite the compile target. However,
+// perhaps we should unify all the page sizes (especially after fastcomp is
+// gone TODO).
+var POSIX_PAGE_SIZE = 16384;
+
 // Receives a function as text, and a function that constructs a modified
 // function, to which we pass the parsed-out name, arguments, and body of the
 // function. Returns the output of that function.
 function modifyFunction(text, func) {
-  var match = text.match(/\s*function\s+([^(]*)?\s*\(([^)]*)\)/);
-  assert(match, 'could not match function ' + text + '.');
-  var name = match[1];
-  var args = match[2];
-  var rest = text.substr(match[0].length);
+  // Match a function with a name.
+  var match = text.match(/^\s*function\s+([^(]*)?\s*\(([^)]*)\)/);
+  var name, args, rest;
+  if (match) {
+    name = match[1];
+    args = match[2];
+    rest = text.substr(match[0].length);
+  } else {
+    // Match a function without a name (we could probably use a single regex
+    // for both, but it would be more complex).
+    match = text.match(/^\s*function\(([^)]*)\)/);
+    assert(match, 'could not match function ' + text + '.');
+    name = '';
+    args = match[1];
+    rest = text.substr(match[0].length);
+  }
   var bodyStart = rest.indexOf('{');
   assert(bodyStart >= 0);
   var bodyEnd = rest.lastIndexOf('}');
@@ -1629,3 +1648,86 @@ function makeRemovedFSAssert(fsName) {
   if (SYSTEM_JS_LIBRARIES.indexOf('library_' + lower + '.js') >= 0) return '';
   return "var " + fsName + " = '" + fsName + " is no longer included by default; build with -l" + lower + ".js';";
 }
+
+// Given an array of elements [elem1,elem2,elem3], returns a string "['elem1','elem2','elem3']"
+function buildStringArray(array) {
+  if (array.length > 0) {
+    return "['" + array.join("','") + "']";
+  } else {
+    return '[]';
+  }
+}
+
+// Generates access to a JS imports scope variable in pthreads worker.js. In MODULARIZE mode these flow into the imports object for the Module.
+// In non-MODULARIZE mode, we can directly access the variables in global scope.
+function makeAsmImportsAccessInPthread(variable) {
+  if (!MINIMAL_RUNTIME) {
+    // Regular runtime uses the name "Module" for both imports and exports.
+    return "Module['" + variable + "']";
+  }
+  if (MODULARIZE) {
+    // MINIMAL_RUNTIME uses 'imports' as the name for the imports object in MODULARIZE builds.
+    return "imports['" + variable + "']";
+  } else {
+    // In non-MODULARIZE builds, can access the imports from global scope.
+    return variable;
+  }
+}
+
+function hasExportedFunction(func) {
+  return Object.keys(EXPORTED_FUNCTIONS).indexOf(func) != -1;
+}
+
+// JS API I64 param handling: if we have BigInt support, the ABI is simple,
+// it is a BigInt. Otherwise, we legalize into pairs of i32s.
+function defineI64Param(name) {
+  if (WASM_BIGINT) {
+    return name + '_bigint';
+  } else {
+    return name + '_low, ' + name + '_high';
+  }
+}
+
+function receiveI64ParamAsI32s(name) {
+  if (WASM_BIGINT) {
+    // TODO: use Xn notation when JS parsers support it (as of April 6 2020,
+    //  * closure compiler is missing support
+    //    https://github.com/google/closure-compiler/issues/3167
+    //  * acorn needs to be upgraded, and to set ecmascript version >= 11
+    //  * terser needs to be upgraded
+    return 'var ' + name + '_low = Number(' + name + '_bigint & BigInt(0xffffffff)) | 0, ' + name + '_high = Number(' + name + '_bigint >> BigInt(32)) | 0;';
+  } else {
+    return '';
+  }
+}
+
+function sendI64Argument(low, high) {
+  if (WASM_BIGINT) {
+    return 'BigInt(low) | (BigInt(high) << BigInt(32))';
+  } else {
+    return low + ', ' + high;
+  }
+}
+
+// Add assertions to catch common errors when using the Promise object we
+// create on Module.ready() and return from MODULARIZE Module() invocations.
+function addReadyPromiseAssertions(promise) {
+  // Warn on someone doing
+  //
+  //  var instance = Module();
+  //  ...
+  //  instance._main();
+  var properties = keys(EXPORTED_FUNCTIONS);
+  // Also warn on onRuntimeInitialized which might be a common pattern with
+  // older MODULARIZE-using codebases.
+  properties.push('onRuntimeInitialized');
+  return properties.map(function(property) {
+    const warningEnding = `${property} on the Promise object, instead of the instance. Use .then() to get called back with the instance, see the MODULARIZE docs in src/settings.js`;
+    return `
+      if (!Object.getOwnPropertyDescriptor(${promise}, '${property}')) {
+        Object.defineProperty(${promise}, '${property}', { configurable: true, get: function() { abort('You are getting ${warningEnding}') } });
+        Object.defineProperty(${promise}, '${property}', { configurable: true, set: function() { abort('You are setting ${warningEnding}') } });
+      }
+    `;
+  }).join('\n');
+}
