@@ -1,13 +1,21 @@
+/**
+ * @license
+ * Copyright 2019 The Emscripten Authors
+ * SPDX-License-Identifier: MIT
+ */
+
 #if STACK_OVERFLOW_CHECK
 // Initializes the stack cookie. Called at the startup of main and at the startup of each thread in pthreads mode.
 function writeStackCookie() {
+#if ASSERTIONS
   assert((STACK_MAX & 3) == 0);
+#endif
 #if WASM_BACKEND
   // The stack grows downwards
-  HEAPU32[(STACK_MAX >> 2)+1] = 0x02135467;
+  HEAPU32[(STACK_MAX >> 2)+1] = 0x2135467;
   HEAPU32[(STACK_MAX >> 2)+2] = 0x89BACDFE;
 #else
-  HEAPU32[(STACK_MAX >> 2)-1] = 0x02135467;
+  HEAPU32[(STACK_MAX >> 2)-1] = 0x2135467;
   HEAPU32[(STACK_MAX >> 2)-2] = 0x89BACDFE;
 #endif
 #if !USE_ASAN
@@ -25,8 +33,8 @@ function checkStackCookie() {
   var cookie1 = HEAPU32[(STACK_MAX >> 2)-1];
   var cookie2 = HEAPU32[(STACK_MAX >> 2)-2];
 #endif
-  if (cookie1 != 0x02135467 || cookie2 != 0x89BACDFE) {
-    abort('Stack overflow! Stack cookie has been overwritten, expected hex dwords 0x89BACDFE and 0x02135467, but received 0x' + cookie2.toString(16) + ' ' + cookie1.toString(16));
+  if (cookie1 != 0x2135467 || cookie2 != 0x89BACDFE) {
+    abort('Stack overflow! Stack cookie has been overwritten, expected hex dwords 0x89BACDFE and 0x2135467, but received 0x' + cookie2.toString(16) + ' ' + cookie1.toString(16));
   }
 #if !USE_ASAN
   // Also test the global address 0 for integrity.
