@@ -9,12 +9,43 @@
 #include <math.h>
 #include <float.h>
 
+#ifndef EMSCRIPTEN_PRINTF_LONG_DOUBLE
+// XXX EMSCRIPTEN - while wasm32 has long double = float128, we don't support
+//                  printing at full precision by default. instead, we lower to
+//                  64-bit double. These macros makes our changes a little less
+//                  invasive.
+typedef double long_double;
+#undef LDBL_TRUE_MIN
+#define LDBL_TRUE_MIN DBL_DENORM_MIN
+#undef LDBL_MIN
+#define LDBL_MIN DBL_MIN
+#undef LDBL_MAX
+#define LDBL_MAX DBL_MAX
+#undef LDBL_EPSILON
+#define LDBL_EPSILON DBL_EPSILON
+#undef LDBL_MANT_DIG
+#define LDBL_MANT_DIG DBL_MANT_DIG
+#undef LDBL_MIN_EXP
+#define LDBL_MIN_EXP DBL_MIN_EXP
+#undef LDBL_MAX_EXP
+#define LDBL_MAX_EXP DBL_MAX_EXP
+#undef LDBL_DIG
+#define LDBL_DIG DBL_DIG
+#undef LDBL_MIN_10_EXP
+#define LDBL_MIN_10_EXP DBL_MIN_10_EXP
+#undef LDBL_MAX_10_EXP
+#define LDBL_MAX_10_EXP DBL_MAX_10_EXP
+#undef frexpl
+#define frexpl(x, exp) frexp(x, exp)
+#else // EMSCRIPTEN_FULL_LONG_DOUBLE_PRINTING
+// XXX EMSCRIPTEN - full long double printing support
+typedef long double long_double;
+#endif
+
 /* Some useful macros */
 
 #define MAX(a,b) ((a)>(b) ? (a) : (b))
 #define MIN(a,b) ((a)<(b) ? (a) : (b))
-#define CONCAT2(x,y) x ## y
-#define CONCAT(x,y) CONCAT2(x,y)
 
 /* Convenient bit representation for modifier flags, which all fall
  * within 31 codepoints of the space character. */
@@ -124,11 +155,20 @@ static const unsigned char states[]['z'-'A'+1] = {
 union arg
 {
 	uintmax_t i;
-	long double f;
+	long_double f;
 	void *p;
 };
 
-static void pop_arg(union arg *arg, int type, va_list *ap)
+// XXX EMSCRIPTEN -  split out long double, so we don't always link in
+//                   long double support for float printf.
+typedef void (*pop_arg_long_double_t)(union arg *arg, va_list *ap);
+
+static void pop_arg_long_double(union arg *arg, va_list *ap)
+{
+  arg->f = va_arg(*ap, long double);
+}
+
+static void pop_arg(union arg *arg, int type, va_list *ap, pop_arg_long_double_t pop_arg_long_double)
 {
 	/* Give the compiler a hint for optimizing the switch. */
 	if ((unsigned)type > MAXSTATE) return;
@@ -154,7 +194,7 @@ static void pop_arg(union arg *arg, int type, va_list *ap)
 	break; case UIPTR:	arg->i = (uintptr_t)va_arg(*ap, void *);
 #endif
 	break; case DBL:	arg->f = va_arg(*ap, double);
-	break; case LDBL:	arg->f = va_arg(*ap, long double);
+	break; case LDBL:	pop_arg_long_double(arg, ap);
 	}
 }
 
@@ -202,10 +242,16 @@ static char *fmt_u(uintmax_t x, char *s)
  * depends on the float.h constants being right. If they are wrong, it
  * may overflow the stack. */
 #if LDBL_MANT_DIG == 53
-typedef char compiler_defines_long_double_incorrectly[9-(int)sizeof(long double)];
+typedef char compiler_defines_long_double_incorrectly[9-(int)sizeof(long_double)];
 #endif
 
-static int fmt_fp(FILE *f, long double y, int w, int p, int fl, int t)
+// XXX EMSCRIPTEN - access fmt_fp indirectly, so that iprintf doesn't 
+//                  get it linked in
+//                  also use a double argument here, as mentioned before,
+//                  we print float128s at double precision 
+typedef int (*fmt_fp_t)(FILE *f, long_double y, int w, int p, int fl, int t);
+
+static int fmt_fp(FILE *f, long_double y, int w, int p, int fl, int t)
 {
 	uint32_t big[(LDBL_MANT_DIG+28)/29 + 1          // mantissa expansion
 		+ (LDBL_MAX_EXP+LDBL_MANT_DIG+28+8)/9]; // exponent expansion
@@ -227,7 +273,7 @@ static int fmt_fp(FILE *f, long double y, int w, int p, int fl, int t)
 
 	if (!isfinite(y)) {
 		char *s = (t&32)?"inf":"INF";
-		if (y!=y) s=(t&32)?"nan":"NAN", pl=0;
+		if (y!=y) s=(t&32)?"nan":"NAN";
 		pad(f, ' ', w, 3+pl, fl&~ZERO_PAD);
 		out(f, prefix, pl);
 		out(f, s, 3);
@@ -239,7 +285,7 @@ static int fmt_fp(FILE *f, long double y, int w, int p, int fl, int t)
 	if (y) e2--;
 
 	if ((t|32)=='a') {
-		long double round = 8.0;
+		long_double round = 8.0;
 		int re;
 
 		if (t&32) prefix += 9;
@@ -343,8 +389,8 @@ static int fmt_fp(FILE *f, long double y, int w, int p, int fl, int t)
 		x = *d % i;
 		/* Are there any significant digits past j? */
 		if (x || d+1!=z) {
-			long double round = CONCAT(0x1p,LDBL_MANT_DIG);
-			long double small;
+			long_double round = 2/LDBL_EPSILON;
+			long_double small;
 			if (*d/i & 1) round += 2;
 			if (x<i/2) small=0x0.8p0;
 			else if (x==i/2 && d+1==z) small=0x1.0p0;
@@ -444,7 +490,9 @@ static int getint(char **s) {
 	return i;
 }
 
-static int printf_core(FILE *f, const char *fmt, va_list *ap, union arg *nl_arg, int *nl_type)
+// XXX EMSCRIPTEN: pass in fmt_fp and pop_arg as a function pointer, so iprintf/__small_printf don't
+//                 force linking in of unnecessary floating-point code.
+static int printf_core(FILE *f, const char *fmt, va_list *ap, union arg *nl_arg, int *nl_type, fmt_fp_t fmt_fp, pop_arg_long_double_t pop_arg_long_double)
 {
 	char *a, *z, *s=(char *)fmt;
 	unsigned l10n=0, fl;
@@ -533,7 +581,7 @@ static int printf_core(FILE *f, const char *fmt, va_list *ap, union arg *nl_arg,
 			if (argpos>=0) return -1;
 		} else {
 			if (argpos>=0) nl_type[argpos]=st, arg=nl_arg[argpos];
-			else if (f) pop_arg(&arg, st, ap);
+			else if (f) pop_arg(&arg, st, ap, pop_arg_long_double);
 			else return 0;
 		}
 
@@ -597,8 +645,10 @@ static int printf_core(FILE *f, const char *fmt, va_list *ap, union arg *nl_arg,
 			*(a=z-(p=1))=arg.i;
 			fl &= ~ZERO_PAD;
 			break;
+#ifndef __EMSCRIPTEN__ // 'm' is a gnu extension, and strerror brings in 2.5K of strings
 		case 'm':
 			if (1) a = strerror(errno); else
+#endif
 		case 's':
 			a = arg.p ? arg.p : "(null)";
 			z = memchr(a, 0, p);
@@ -646,13 +696,15 @@ static int printf_core(FILE *f, const char *fmt, va_list *ap, union arg *nl_arg,
 	if (!l10n) return 0;
 
 	for (i=1; i<=NL_ARGMAX && nl_type[i]; i++)
-		pop_arg(nl_arg+i, nl_type[i], ap);
+		pop_arg(nl_arg+i, nl_type[i], ap, pop_arg_long_double);
 	for (; i<=NL_ARGMAX && !nl_type[i]; i++);
 	if (i<=NL_ARGMAX) return -1;
 	return 1;
 }
 
-int vfprintf(FILE *restrict f, const char *restrict fmt, va_list ap)
+// XXX EMSCRIPTEN: pass in fmt_fp and pop_arg as a function pointer, so iprintf/__small_printf don't
+//                 force linking in of floating-point code.
+int __vfprintf_internal(FILE *restrict f, const char *restrict fmt, va_list ap, fmt_fp_t fmt_fp, pop_arg_long_double_t pop_arg_long_double)
 {
 	va_list ap2;
 	int nl_type[NL_ARGMAX+1] = {0};
@@ -663,7 +715,7 @@ int vfprintf(FILE *restrict f, const char *restrict fmt, va_list ap)
 
 	/* the copy allows passing va_list* even if va_list is an array */
 	va_copy(ap2, ap);
-	if (printf_core(0, fmt, &ap2, nl_arg, nl_type) < 0) {
+	if (printf_core(0, fmt, &ap2, nl_arg, nl_type, fmt_fp, pop_arg_long_double) < 0) {
 		va_end(ap2);
 		return -1;
 	}
@@ -677,7 +729,7 @@ int vfprintf(FILE *restrict f, const char *restrict fmt, va_list ap)
 		f->buf_size = sizeof internal_buf;
 		f->wend = internal_buf + sizeof internal_buf;
 	}
-	ret = printf_core(f, fmt, &ap2, nl_arg, nl_type);
+	ret = printf_core(f, fmt, &ap2, nl_arg, nl_type, fmt_fp, pop_arg_long_double);
 	if (saved_buf) {
 		f->write(f, 0, 0);
 		if (!f->wpos) ret = -1;
@@ -691,3 +743,20 @@ int vfprintf(FILE *restrict f, const char *restrict fmt, va_list ap)
 	va_end(ap2);
 	return ret;
 }
+
+int vfprintf(FILE *restrict f, const char *restrict fmt, va_list ap)
+{
+	return __vfprintf_internal(f, fmt, ap, fmt_fp, pop_arg_long_double);
+}
+
+// XXX EMSCRIPTEN
+int vfiprintf(FILE *restrict f, const char *restrict fmt, va_list ap)
+{
+	return __vfprintf_internal(f, fmt, ap, NULL, NULL);
+}
+
+// XXX EMSCRIPTEN
+int __small_vfprintf(FILE *restrict f, const char *restrict fmt, va_list ap)
+{
+	return __vfprintf_internal(f, fmt, ap, fmt_fp, NULL);
+}
