@@ -1,10 +1,17 @@
-from toolchain_profiler import ToolchainProfiler
-import time, os, sys, logging
-from subprocess import Popen, PIPE, STDOUT
+# Copyright 2013 The Emscripten Authors.  All rights reserved.
+# Emscripten is available under two separate licenses, the MIT license and the
+# University of Illinois/NCSA Open Source License.  Both these licenses can be
+# found in the LICENSE file.
 
-TRACK_PROCESS_SPAWNS = True if (os.getenv('EM_BUILD_VERBOSE') and int(os.getenv('EM_BUILD_VERBOSE')) >= 3) else False
+import logging
+import os
+import sys
+import time
+from subprocess import Popen, PIPE, CalledProcessError
+
+TRACK_PROCESS_SPAWNS = int(os.getenv('EM_BUILD_VERBOSE', '0')) >= 3
 WORKING_ENGINES = {} # Holds all configured engines and whether they work: maps path -> True/False
-__rootpath__ = os.path.dirname(os.path.dirname(__file__))
+
 
 def timeout_run(proc, timeout=None, note='unnamed process', full_output=False, note_args=[], throw_on_failure=True):
   start = time.time()
@@ -14,14 +21,15 @@ def timeout_run(proc, timeout=None, note='unnamed process', full_output=False, n
     if proc.poll() is None:
       proc.kill() # XXX bug: killing emscripten.py does not kill it's child process!
       raise Exception("Timed out: " + note)
-  out = proc.communicate()
-  out = map(lambda o: '' if o is None else o, out)
+  stdout, stderr = proc.communicate()
+  out = ['' if o is None else o for o in (stdout, stderr)]
   if throw_on_failure and proc.returncode != 0:
-    raise Exception('Subprocess "' + ' '.join(note_args) + '" failed with exit code ' + str(proc.returncode) + '!')
+    raise CalledProcessError(proc.returncode, ' '.join(note_args), stdout, stderr)
   if TRACK_PROCESS_SPAWNS:
     logging.info('Process ' + str(proc.pid) + ' finished after ' + str(time.time() - start) + ' seconds. Exit code: ' + str(proc.returncode))
   return '\n'.join(out) if full_output else out[0]
 
+
 def make_command(filename, engine=None, args=[]):
   if type(engine) is not list:
     engine = [engine]
@@ -33,15 +41,24 @@ def make_command(filename, engine=None, args=[]):
   #
   # Check only the last part of the engine path to ensure we don't accidentally
   # label a path to nodejs containing a 'd8' as spidermonkey instead.
-  jsengine = os.path.split(engine[0])[-1]
+  jsengine = os.path.basename(engine[0])
   # Use "'d8' in" because the name can vary, e.g. d8_g, d8, etc.
-  is_d8 = 'd8' in jsengine
+  is_d8 = 'd8' in jsengine or 'v8' in jsengine
+  is_jsc = 'jsc' in jsengine
+  is_wasmer = 'wasmer' in jsengine
+  is_wasmtime = 'wasmtime' in jsengine
   # Disable true async compilation (async apis will in fact be synchronous) for now
   # due to https://bugs.chromium.org/p/v8/issues/detail?id=6263
   shell_option_flags = ['--no-wasm-async-compilation'] if is_d8 else []
+  command_flags = []
+  if is_wasmer:
+    command_flags += ['run']
+  if is_wasmer or is_wasmtime:
+    # in a wasm runtime, run the wasm, not the js
+    filename = shared.unsuffixed(filename) + '.wasm'
   # Separates engine flags from script flags
-  flag_separator = ['--'] if is_d8 or 'jsc' in jsengine else []
-  return engine + [filename] + shell_option_flags + flag_separator + args
+  flag_separator = ['--'] if is_d8 or is_jsc else []
+  return engine + command_flags + [filename] + shell_option_flags + flag_separator + args
 
 
 def check_engine(engine):
@@ -54,9 +71,11 @@ def check_engine(engine):
     return WORKING_ENGINES[engine_path]
   try:
     logging.debug('Checking JS engine %s' % engine)
-    if 'hello, world!' in run_js(os.path.join(__rootpath__, 'src', 'hello_world.js'), engine, skip_check=True):
+    output = run_js(shared.path_from_root('src', 'hello_world.js'), engine,
+                    skip_check=True)
+    if 'hello, world!' in output:
       WORKING_ENGINES[engine_path] = True
-  except Exception, e:
+  except Exception as e:
     logging.info('Checking JS engine %s failed. Check your config file. Details: %s' % (str(engine), str(e)))
     WORKING_ENGINES[engine_path] = False
   return WORKING_ENGINES[engine_path]
@@ -71,8 +90,23 @@ def require_engine(engine):
     sys.exit(1)
 
 
+def run_js_tool(filename, engine, jsargs=[], *args, **kw):
+  """Execute a javascript tool.
+
+  This is used by emcc to run parts of the build process that are written
+  implemented in javascript.
+  """
+  command = make_command(filename, engine, jsargs)
+  return shared.check_call(command, *args, **kw).stdout
+
+
+# TODO(sbc): Move to into test directory
 def run_js(filename, engine=None, args=[], check_timeout=False, stdin=None, stdout=PIPE, stderr=None, cwd=None,
-           full_output=False, assert_returncode=0, error_limit=-1, skip_check=False):
+           full_output=False, assert_returncode=0, skip_check=False):
+  """Execute javascript code generated by tests, with possible timeout."""
+  if not os.path.exists(filename):
+    raise Exception('JavaScript file not found: ' + filename)
+
   #  # code to serialize out the test suite files
   #  # XXX make sure to disable memory init files, and clear out the base_dir. you may also need to manually grab e.g. paper.pdf.js from a run of test_poppler
   #  import shutil, json
@@ -94,23 +128,21 @@ def run_js(filename, engine=None, args=[], check_timeout=False, stdin=None, stdo
 
   command = make_command(filename, engine, args)
   try:
-    if cwd is not None: os.environ['EMCC_BUILD_DIR'] = os.getcwd()
     proc = Popen(
         command,
         stdin=stdin,
         stdout=stdout,
         stderr=stderr,
-        cwd=cwd)
-  except Exception, e:
+        cwd=cwd,
+        universal_newlines=True)
+  except Exception:
     # the failure may be because the engine is not present. show the proper
     # error in that case
     if not skip_check:
       require_engine(engine)
     # if we got here, then require_engine succeeded, so we can raise the original error
-    raise e
-  finally:
-    if cwd is not None: del os.environ['EMCC_BUILD_DIR']
-  timeout = 15*60 if check_timeout else None
+    raise
+  timeout = 15 * 60 if check_timeout else None
   if TRACK_PROCESS_SPAWNS:
     logging.info('Blocking on process ' + str(proc.pid) + ': ' + str(command) + (' for ' + str(timeout) + ' seconds' if timeout else ' until it finishes.'))
   try:
@@ -120,13 +152,20 @@ def run_js(filename, engine=None, args=[], check_timeout=False, stdin=None, stdo
       'Execution',
       full_output=full_output,
       throw_on_failure=False)
-  except Exception, e:
+  except Exception:
     # the failure may be because the engine does not work. show the proper
     # error in that case
     if not skip_check:
       require_engine(engine)
     # if we got here, then require_engine succeeded, so we can raise the original error
-    raise e
+    raise
   if assert_returncode is not None and proc.returncode is not assert_returncode:
-    raise Exception('Expected the command ' + str(command) + ' to finish with return code ' + str(assert_returncode) + ', but it returned with code ' + str(proc.returncode) + ' instead! Output: ' + str(ret)[:error_limit])
+    raise CalledProcessError(proc.returncode, ' '.join(command), str(ret))
   return ret
+
+
+try:
+  from . import shared
+except ImportError:
+  # Python 2 circular import compatibility
+  import shared
